#include <stdio.h>
#include <stdlib.h>

typedef struct node {
  double val;
  struct node *next;
  struct node *prev;
} node;

typedef struct container {
  node *first;
  node *last;
} container;


container *make_sequence(int n)
{
  container *c;
  node *fst = NULL, *nd, *prev = NULL;
  int i;

  c = malloc(sizeof(container));

  for (i = 0; i < n; i++) {
    nd = malloc(sizeof(node));
    nd->val = i;
    nd->prev = prev;
    nd->next = NULL;

    if (prev)
      prev->next = nd;
    else
      c->first = nd;

    prev = nd;
  }

  c->last = prev;

  return c;
}

void change_list_to_defeat_optimization(container *c)
{
  if (c->first)
    c->first->val = -c->first->val;
}

static double sum_list(container *c)
{
  node *nd;
  double sum = 0.0;

  for (nd = c->first; nd; nd = nd->next) {
    sum += nd->val;
  }

  return sum;
}

static double sum_list2(container *c)
{
  node *nd, *nd2;
  double sum = 0.0;

  for (nd = c->first, nd2 = c->last; nd != nd2; nd = nd->next, nd2 = nd2->prev) {
    sum += nd->val;
    sum += nd2->val;
    if (nd->next == nd2)
      break;
  }

  return sum;
}

static double sum_list3(container *c)
{
  node *nd, *nd2;
  double sum = 0.0, sum2 = 0.0;

  for (nd = c->first, nd2 = c->last; nd != nd2; nd = nd->next, nd2 = nd2->prev) {
    sum += nd->val;
    sum2 += nd2->val;
    if (nd->next == nd2)
      break;
  }

  return sum + sum2;
}

static inline unsigned long long get_ticks(void)
{
  unsigned int lo, hi;
  // RDTSC copies contents of 64-bit TSC into EDX:EAX
  asm volatile("rdtsc" : "=a" (lo), "=d" (hi));
  return (unsigned long long)hi << 32 | lo;
}

#define LENGTH 1000
#define ITERS  100000

int main(int argc, char **argv)
{
  container *c;
  int i, sel;
  double sum = 0.0;
  unsigned long long start_ticks, end_ticks;

  sel = ((argc > 1) ? atoi(argv[1]) : 1);

  c = make_sequence(LENGTH);

  start_ticks = get_ticks();

  switch (sel) {
  case 2:
    for (i = 0; i < ITERS; i++) {
      sum += sum_list2(c);
      change_list_to_defeat_optimization(c);
    }
    break;
  case 3:
    for (i = 0; i < ITERS; i++) {
      sum += sum_list3(c);
      change_list_to_defeat_optimization(c);
    }
    break;
  default:
    for (i = 0; i < ITERS; i++) {
      sum += sum_list(c);
      change_list_to_defeat_optimization(c);
    }
  }

  end_ticks = get_ticks();

  printf("%f, %f\n",
         sum,
         ((double)(end_ticks - start_ticks) / (LENGTH * ITERS)) * (4.0 / 3.6));

  return 0;
}
